/*
 * (c) 2021, SWD Embedded Systems Limited, http://www.kpda.ru
 */


#include <stddef.h>       /* null, size, uint */
#include <sys/mman.h>
#include <hw/inout.h>
#include <sys/neutrino.h>
#include <hw/spi-master.h>
#include "s25-spi.h"
#include <sys/slog.h>
#include <sys/slogcodes.h>


#define SPI_SET_ADDRESS( a, b )         \
    do {                                \
        uint8_t*    _b = (void *)(b);   \
        _b[1] = (((a) >> 16) & 0xFF);   \
        _b[2] = (((a) >>  8) & 0xFF);   \
        _b[3] = (((a) >>  0) & 0xFF);   \
    } while ( 0 )
#define WAITTIME            30000
#define _SLOGC_FS_FLASH     __C( _SLOGC_FS, 2000 )


void print_data( unsigned char *buffer, unsigned len )
{
    int     idx;

    printf( "Data:" );
    for ( idx = 0; idx < len; idx++ )
    {
        if ( (idx & 0x1F) == 0 )
            printf( "\n" );
        printf( "%02x ", buffer[idx] );
    }
    printf( "\n" );
}


/*
 * ------------------------------
 * LOWLEVEL FUNCTIONS
 * ------------------------------
 */

int s25_spi_init( void *vbase )
{
    s25_spi_t   *spi = (s25_spi_t *)vbase;
    char        devname[20];
    spi_cfg_t   cfg;

    cfg.mode = SPI_MODE_INIT_VAL;

    snprintf( devname, 20, "/dev/spi%d", spi->devnum );
    if ( (spi->fd = spi_open( devname )) < 0 )
    {
        slogf( _SLOGC_FS_FLASH, _SLOG_ERROR, "SPI open failed, device = %s, fd = %d errno = %d  %s", devname, spi->fd, errno, strerror( errno ) );
        return (1);
    }
    spi_setcfg( spi->fd, spi->cs | SPI_DEV_DEFAULT, &cfg );
    return (0);
}

int s25_spi_transfer( void *vbase, int line, void *tx_, int tx_len, void *rx_, int rx_len )
{
    s25_spi_t       *spi     = (s25_spi_t*)vbase;
    int             i,
                    status   = 0,
                    cmd_len  = (tx_len < 4 ? tx_len : 4),
                    xfer_len = tx_len + rx_len;
    unsigned char   bf1[xfer_len + 4],
                    bf2[xfer_len + 4],
                    *tx_buf  = bf1,
                    *rx_buf  = bf2;

    memcpy( tx_buf, tx_, tx_len );
    for ( i = tx_len; i < xfer_len; i++ )
        tx_buf[i] = 0xFF;
    i = spi_xchange( spi->fd, spi->cs | SPI_DEV_DEFAULT, (uint8_t *)tx_buf, (uint8_t *)rx_buf, xfer_len );
    memcpy( rx_, rx_buf + tx_len, rx_len );

#if 0
    printf( "SPI xfer, tx_len = %d, rx_len = %d, rsize = %d, fd = %d\n", tx_len, rx_len, i, spi->fd );
    print_data( tx_buf, xfer_len );
    print_data( rx_buf, xfer_len );
#endif

    return i-cmd_len;
}

int s25_spi_exec( void *vbase, int line, uint8_t cmd_op, uint32_t address, uint8_t *buffer, uint32_t buf_len )
{
    uint8_t     cmd[4 + SPI_MAX_SIZE];  /* largest command takes 4 bytes, the data block occupies SPI_MAX_SIZE bytes */
    uint8_t     *in,
                *out;
    uint16_t    in_len,
                out_len;
    size_t      size;
    int         i;

    cmd[0] = cmd_op;                /* Save the SPI flash instruction. */
    switch ( cmd_op )               /* Prepare arguments for the SPI transaction. */
    {
        case SPI_FLASH_RDID:        /* Read identification. */
            in      = cmd;
            in_len  = 1;
            out     = buffer;
            out_len = (buf_len > JEDEC_DATA) ? JEDEC_DATA : buf_len;
            break;

        case SPI_FLASH_FAST_READ:   /* Read Data Bytes */
        case SPI_FLASH_READ:        /* Read Data Bytes */
            in      = cmd;
            in_len  = 4;
            out     = buffer;
            out_len = (buf_len > SPI_MAX_SIZE) ? SPI_MAX_SIZE : buf_len;
            SPI_SET_ADDRESS( address, cmd );
            break;

        case SPI_FLASH_RDSR:        /* Read Status Register */
            in      = cmd;
            in_len  = 1;
            out     = buffer;
            out_len = buf_len;
            break;

        case SPI_FLASH_WRSR:        /* Write Status Register */
            in      = cmd;
            in_len  = 2;
            out     = 0;
            out_len = 0;
            cmd[1]  = buffer[0];
            break;

        case SPI_FLASH_SSE:         /* SubSector Erase */
        case SPI_FLASH_SE:          /* Sector Erase */
            in      = cmd;
            in_len  = 4;
            out     = 0;
            out_len = 0;
            SPI_SET_ADDRESS( address, cmd );
            break;

        case SPI_FLASH_WRDI:        /* Write Disable */
        case SPI_FLASH_WREN:        /* Write Enable */
        case SPI_FLASH_BE:          /* Bulk Erase */
            in      = cmd;
            in_len  = 1;
            out     = 0;
            out_len = 0;
            break;

        case SPI_FLASH_PP:          /* Page Program */
            size    = (buf_len > PAGE_SIZE) ? PAGE_SIZE : buf_len;
            in      = cmd;
            in_len  = 4 + size;
            out     = 0;
            out_len = 0;
            SPI_SET_ADDRESS( address, cmd );
            for ( i = 0; i < size; i++ )
                cmd[i + 4] = buffer[i];
            break;

        default:
            /* printf( "%s: SPI instruction %d is not provided\n", __FUNCTION__, cmd_op ); */
            return (-1);
    }

    /* Execute the SPI transaction */
    return s25_spi_transfer( vbase, line, in, in_len, out, out_len );
}


/*
 * ------------------------------
 * HIGH LEVEL FUNCTIONS
 * ------------------------------
 */

int s25_spi_id( void *vbase, void *data, uint32_t size )
{
    return s25_spi_exec( vbase, 0, SPI_FLASH_RDID, 0, data, size );
}

int s25_spi_status( void *vbase, void *status )
{
    return s25_spi_exec( vbase, 0, SPI_FLASH_RDSR, 0, status, 1 );
}

int s25_spi_wren( void *vbase )
{
    uint8_t     status;
    uint32_t    timeout = WAITTIME;

    s25_spi_exec( vbase, 0, SPI_FLASH_WREN, 0, 0, 0 );
    s25_spi_exec( vbase, 0, SPI_FLASH_RDSR, 0, &status, 1 );

     do {
        if ( (timeout--) == 0 )
        {
            slogf( _SLOGC_FS_FLASH, _SLOG_ERROR, "Error: s25_spi_wren(WIP) timeout" );
            return (-1);
        }

        s25_spi_exec( vbase, 0, SPI_FLASH_RDSR, 0, &status, 1 );
    } while ( status & SPI_FLASH_SR_WIP );

    timeout = WAITTIME;
    do {
        if ( (timeout--) == 0 )
        {
            slogf( _SLOGC_FS_FLASH, _SLOG_ERROR, "Error: s25_spi_wren timeout" );
            return (-1);
        }
        s25_spi_exec( vbase, 0, SPI_FLASH_RDSR, 0, &status, 1 );
    } while ( !(status & SPI_FLASH_SR_WEL) );

    return (status & SPI_FLASH_SR_WEL) ? 0 : -1;
}

int s25_spi_wait( void *vbase )
{
    uint8_t     status;
    uint32_t    timeout = WAITTIME;

    do {
        if ( (timeout--) == 0 )
        {
            slogf( _SLOGC_FS_FLASH, _SLOG_ERROR, "Error: s25_spi_wait timeout" );
            return (-1);
        }

        nanospin_ns( 300000 );
        s25_spi_exec( vbase, 0, SPI_FLASH_RDSR, 0, &status, 1 );
    } while ( status & SPI_FLASH_SR_WIP );

    return (0);
}

int s25_spi_erase( void *vbase, int adr, int mode )
{
    int     err = s25_spi_wren( vbase );

    if ( err )
        return err;

    if ( mode )
        err = s25_spi_exec( vbase, 0, SPI_FLASH_SSE, adr, 0, 0 );
    else
        err = s25_spi_exec( vbase, 0, SPI_FLASH_SE, adr, 0, 0 );

    if ( err < 0 )
        return err;

    err = s25_spi_wait( vbase );
    if ( err )
        return err;

    return (0);
}

int s25_spi_write_page( void *vbase, int adr, void *data, uint32_t size )
{
    int     err,
            wr_len;

    err = s25_spi_wren( vbase );
    if ( err )
    {
        slogf( _SLOGC_FS_FLASH, _SLOG_ERROR, "Error: SPI wren" );
        return err;
    }

    wr_len = s25_spi_exec( vbase, 0, SPI_FLASH_PP, adr, data, size );
    if ( wr_len != size )
    {
        slogf( _SLOGC_FS_FLASH, _SLOG_ERROR, "Error: SPI write" );
        return wr_len;
    }

    err = s25_spi_wait( vbase );
    if ( err )
    {
        slogf( _SLOGC_FS_FLASH, _SLOG_ERROR, "Error: SPI wait" );
        return err;
    }

    return wr_len;
}

/* Flash writes must be aligned to PAGE_SIZE boundary */
int aligned_flash_write( void *vbase, unsigned offset, unsigned char *buffer, unsigned size )
{
    int         overflow        = offset % PAGE_SIZE + size;
    unsigned    to_upper_border = size - (offset + size) % PAGE_SIZE;

    if ( overflow > PAGE_SIZE )
    {
        /* Write bytes that fit in current page */
        if ( s25_spi_write_page( vbase, offset, buffer, to_upper_border ) != to_upper_border )
            return (-1);

        /* Write left bytes in next page */
        if ( s25_spi_write_page( vbase, offset + to_upper_border, buffer + to_upper_border, (offset + size) % PAGE_SIZE) != (offset + size) % PAGE_SIZE )
            return (-1);
    } else {
        if ( s25_spi_write_page( vbase, offset, buffer, size ) != size )
            return (-1);
    }
    return (0);
}

int s25_spi_write( void *vbase, int adr, void *buffer, int size )
{
    unsigned    current_offset = adr,
                size_left      = size;

    for ( current_offset = adr; current_offset < adr + size; current_offset += PAGE_SIZE )
    {
        if ( size_left < PAGE_SIZE )
        {
            if ( aligned_flash_write( vbase, current_offset, buffer + (size - size_left), size_left ) != 0 )
                slogf( _SLOGC_FS_FLASH, _SLOG_ERROR, "%s %d: write status failed for offset 0x%x",
                       __func__, __LINE__, current_offset );
        } else
            if ( aligned_flash_write( vbase, current_offset, buffer + (size - size_left), PAGE_SIZE ) != 0 )
                slogf( _SLOGC_FS_FLASH, _SLOG_ERROR, "%s %d: write status failed for offset 0x%x",
                       __func__, __LINE__, current_offset );

        size_left -= PAGE_SIZE;
    }
    return size;
}

int s25_spi_read_page( void *vbase, int adr, void *data, uint32_t size )
{
    int     err,
            rcvd_len;

    rcvd_len = s25_spi_exec( vbase, 0, SPI_FLASH_READ, adr, data, size );
    if ( rcvd_len != size )
        return rcvd_len;

    err = s25_spi_wait( vbase );
    if ( err )
    {
        slogf( _SLOGC_FS_FLASH, _SLOG_ERROR, "Error: SPI wait" );
        return err;
    }

    return rcvd_len;
}

int s25_spi_read( void *vbase, int adr, void *sector, int size )
{
    int         rcvd_len = 0,
                ret;
    uint32_t    cnt      = size;
    uint8_t     *p       = sector;

    while ( cnt )
    {
        size = (cnt < SPI_MAX_SIZE) ? cnt : SPI_MAX_SIZE;
        ret  = s25_spi_read_page( vbase, adr, p, size );

        if ( ret < size )
        {
            ret = s25_spi_read_page( vbase, adr, p, size );
            if ( ret < size )
                return (rcvd_len + ret);
        }

        p   += ret;
        adr += ret;
        cnt -= ret;

        rcvd_len += ret;
    }

    return rcvd_len;
}

void s25_spi_close( void *vbase )
{
    s25_spi_t   *spi = (s25_spi_t *)vbase;

    spi_close( spi->fd );
    /*free( spi );*/
}
