/*
 * (c) 2021, SWD Embedded Systems Limited, http://www.kpda.ru
 */

#ifndef __S25_SPI_H_
#define __S25_SPI_H__


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <inttypes.h>
#include <fcntl.h>
#include <sys/types.h>
#include <stddef.h>
#include "spi_regs.h"
#include "spi_boot.h"


typedef struct s25_spi_s {
    int32_t     fd;
    int8_t      devnum;
    int8_t      cs;
} s25_spi_t;

#define DEVICE_IDENT        "M25"

#define PAGE_SIZE           256     /* bytes */
#define SECTOR_SIZE         256     /* number of pages in one sector */
#define TOTAL_SIZE          256     /* sectors */
#define ADDR_LEN            3       /* 24 bit address */
#define CMD_LEN             1       /* one byte commands */

#define TOTAL_SIZE_BYTES    (TOTAL_SIZE * SECTOR_SIZE * PAGE_SIZE)

#define SPI_MODE            0
#define DEVICE_ID           0x18
#define MAN_ID              0x20
#define NUM_ERASE_UNITS     256     /* Number of sectors, this is the erase unit */
#define PWR2_SIZE_UNIT      16

/* debug config */
#define SPI_CHECK_SAVEENV   0       /* check all sectors after u-boot cmd:saveenv */
#define SPI_LOWLEVEL_DEBUG  0       /* show lowlewel spi debug print */

/* constants  */
#define SPI_FLASH_SIZE      0x1000000
#define SPI_QNX_SIZE        0xf00000
#define SPI_QNX_OFFSET      0x100000
#define SPI_SECTOR_SIZE     0x10000
#define SPI_SUBSECTOR_SIZE  0x10000

#define SPI_MAX_SIZE        0x8000
#define SPI_BAUDR_SCKDV     6       /* default baud rate and thresholds. */
#define JEDEC_DATA          20
#define MAX_SPI_TRIES       10000
#define SPI_FIFO_SIZE       (CONFIG_SPI_FIFO_SIZE / sizeof( uint16_t ))
#define SPI_TH_TX           (SPI_FIFO_SIZE / sizeof( uint16_t ))        /* fifo threshold */
#define SPI_TH_RX           (SPI_FIFO_SIZE / sizeof( uint16_t ))
#define BOOT_SPI_PORT       0

#define ERASE_SECTOR        0
#define ERASE_SUBSECTOR     1

#define SPI_MODE_INIT_VAL   0x80000403

/* SPI registers definitions */
/* SPI_SSIENR */
#define SSIENR_SSI_DE       0
#define SSIENR_SSI_EN       1

/* CTRL 0 */
/* Slave output enable. */
#define SPI_SLV_OE_OFFSET   10
#define SPI_SLV_MASK        (0x1 << SPI_SLV_OE_OFFSET)
#define SPI_SLV_OE          1

/* Transfer mode bits */
#define SPI_TMOD_OFFSET     8
#define SPI_TMOD_MASK       (0x3 << SPI_TMOD_OFFSET)
#define SPI_TMOD_TR         0x0     /* transmit and receive */
#define SPI_TMOD_TO         0x1     /* transmit only */
#define SPI_TMOD_RO         0x2     /* receive only */
#define SPI_TMOD_EPROMREAD  0x3     /* EEPROM read */

/* Serial Clock Polarity. */
#define SPI_SCPOL_OFFSET    7
#define SPI_SCPOL_MASK      (0x3 << SPI_SCPOL_OFFSET)
#define SPI_SCPOL_LOW       0
#define SPI_SCPOL_HIGH      1

/* Serial Clock Phase. */
#define SPI_SCPH_OFFSET     6
#define SPI_SCPH_MASK       (0x3 << SPI_SCPH_OFFSET)
#define SPI_SCPH_MIDDLE     0
#define SPI_SCPH_START      1

/* Frame format. */
#define SPI_FRF_OFFSET      4
#define SPI_FRF_MASK        (0x3 << SPI_FRF_OFFSET)
#define SPI_FRF_SPI         0x0
#define SPI_FRF_SSP         0x1
#define SPI_FRF_MICROWIRE   0x2
#define SPI_FRF_RESV        0x3

/* Data Frame Size */
#define SPI_DFS_OFFSET      0
#define SPI_DFS_MASK        (0x3 << SPI_DFS_OFFSET)
#define SPI_DFS( x )        (x - 1)

/* Status register bits. */
#define SPI_SR_DCOL         (1 << 6)    /* Data Collision Error */
#define SPI_SR_TXE          (1 << 5)    /* Transmition Error. */
#define SPI_SR_RFF          (1 << 4)    /* Receive FIFO Full */
#define SPI_SR_RFNE         (1 << 3)    /* Receive FIFO Not Empty */
#define SPI_SR_TFE          (1 << 2)    /* Transmit FIFO Empty */
#define SPI_SR_TFNF         (1 << 1)    /* Transmit FIFO Not Full */
#define SPI_SR_BUSY         (1 << 0)    /* SSI Busy Flag. */

/* Define the SPI bits of the status register. */
#define SPI_FLASH_SR_WIP    (1 << 0)    /* Write In Progress */
#define SPI_FLASH_SR_WEL    (1 << 1)    /* Write Enable Latch */

/* Flag Status Register */
#define SPI_FLAG_PE         (1 << 7)    /* P/E Controller (not WIP) */
#define SPI_FLAG_ER_SUSPEND (1 << 6)    /* Erase Suspend */
#define SPI_FLAG_ERASE      (1 << 5)    /* Erase */
#define SPI_FLAG_PROGRAM    (1 << 4)    /* Program */
#define SPI_FLAG_VPP        (1 << 3)    /* VPP */
#define SPI_FLAG_PR_SUSPEND (1 << 2)    /* Program Suspend */
#define SPI_FLAG_PROTECTION (1 << 1)    /* Protection */
#define SPI_FLAG_           (1 << 0)    /* RESERVED */


/* SPI Flash commands */
#define SPI_FLASH_RDID      0x9F        /* (0, 1-20) Read identification. */
#define SPI_FLASH_READ      0x03        /* (3, 1-∞ ) Read Data Bytes */
#define SPI_FLASH_FAST_READ 0x0B        /* (3, 1-∞ ) Read Data Bytes */
#define SPI_FLASH_WREN      0x06        /* (0, 0   ) Write Enable */
#define SPI_FLASH_WRDI      0x04        /* (0, 0   ) Write Disable */
#define SPI_FLASH_PP        0x02        /* (3, 256 ) Page Program */
#define SPI_FLASH_SSE       0x20        /* (3, 0   ) SubSector Erase */
#define SPI_FLASH_SE        0xD8        /* (3, 0   ) Sector Erase */
#define SPI_FLASH_RDSR      0x05        /* (0, 1   ) Read Status Register */
#define SPI_FLASH_WRSR      0x01        /* (0, 1-∞ ) Write Status Register */
#define SPI_FLASH_RDLR      0xE8        /* (3, 1-∞ ) Read Lock Register */
#define SPI_FLASH_WRLR      0xE5        /* (3, 1   ) Write Lock Register */
#define SPI_FLASH_RFSR      0x70        /* (1 to ∞)  Read Flag Status Register */
#define SPI_FLASH_CLFSR     0x50        /* (0) Clear Flag Status Register */
#define SPI_FLASH_BE        0xC7        /* (0) Bulk Erase */
#define SPI_FLASH_RSTEN     0x66        /* Reset Enable */
#define SPI_FLASH_RST       0x99        /* Reset Memory */
/* RDNVCR Read NV Configuration Register B5h (2) */
/* WRNVCR Write NV Configuration Register B1h (2) */
/* RDVCR Read Volatile Configuration Register 85h (1 to ∞) */
/* WRVCR Write Volatile Configuration Register 81h (1) */
/* RDVECR Read Volatile Enhanced Configuration Register 65h (1 to ∞) */
/* WRVECR Write Volatile Enhanced Configuration Register 61h (1) */


/* SPI FUNCTIONS */
int s25_spi_init( void *vbase );
int s25_spi_id( void *vbase, void *data, uint32_t size );
int s25_spi_wren( void *vbase );
int s25_spi_wait( void *vbase );
int s25_spi_erase( void *vbase, int adr, int mode );
int s25_spi_status( void *vbase, void *status );
int s25_spi_read_page( void *vbase, int adr, void *data, uint32_t size );
int s25_spi_write_page( void *vbase, int adr, void *data, uint32_t size );
int s25_spi_read( void *vbase, int adr, void *sector, int size );
int s25_spi_write( void *vbase, int adr, void *sector, int size );
void s25_spi_close( void *vbase );
int iswriting( void *vbase );


#endif  /* __S25_SPI_H_ */
